from abc import ABCMeta, abstractmethod
import logging
import os
logger = logging.getLogger(os.path.basename(__file__))


def io_connector(image_path, label_path, *args, **kwds):
    '''
    Returns either a TiffConnector or an NapariConnector, depending on
    input files.

    Parameters
    ----------
    image_path : str or list of str
        Either wildcard or list of paths to pixel data in tiff format.
    label_path : str of list of str
        Either wildcard or list of paths to pixel data in tiff format
        returns a TiffConnector. If a path to a single Ilastik ilp
        file is given, an NapariConnector is returned.

    Returns
    -------
    Connector
        Either NapariConnector or TiffConnector

    See Also
    --------
    yapic_io.tiff_connector.TiffConnector
    yapic_io.napari_connector.NapariConnector
    '''
    from yapic_io.tiff_connector import TiffConnector
    from yapic_io.napari_connector import NapariConnector

    if label_path.endswith('.h5'):
        logger.info('Napari project file detected')
        return NapariConnector(image_path, label_path)
    else:
        logger.info('Tiff files detected.')
        return TiffConnector(image_path, label_path, *args, **kwds)


class Connector(metaclass=ABCMeta):
    '''
    Interface to pixel and label data source for classifier
    training and prediction.

    - Connects to a fixed collection of images and corresponding labels
    - Provides methods for getting sizes of all images
    - Provides methods for getting image subsets and label coordinates.


    Prerequisites for the Image data:

    - Images are 4D datasets with following dimensions: (channel, z, x, y)
    - All images must have the same nr of channels but can vary in z, x, and y.
    - Time series are not supported. However, time frames can be modeled as
      individual images.
    - Images with lower dimensions can be modeled. E.g. grayscale single plane
      images
      have the shape (1, 1, width, height).

    The Connector methods are used by the Dataset class.
    '''

    def __init__(self):
        '''
        In polling mode, the dataset will repeatedly fetch
        tiles for training (using get_tile) until labels of a
        certain value are present.
        This is efficient in following situation:
        - relatively small images
        - dense labelling

        For very large images with sparse labelling,
        the by_label_index mode is recommended (implemented in
        Coordinate_connector class)
        '''


    @abstractmethod
    def image_count(self):
        '''
        Get number of images.

        Returns
        -------
        int
            Number of images in the dataset.
        '''
        pass

    @abstractmethod
    def label_count_for_image(self, image_nr):
        '''
        Returns for each label value the number of labels for this image

        Parameters
        ----------
        image_nr : int
            Index of image.

        Returns
        -------
        dict
            Counts per label and per image.
        '''
        pass

    @abstractmethod
    def get_tile(self, image_nr=None, pos=None, size=None):
        '''
        Get 4D subsection of an image.

        Parameters
        ----------
        image_nr : int
            Index of image.
        pos : (channel, zslice, x, y)
            Upper left position of subsection

        Returns
        -------
        numpy.ndarray
            4D subsection of image as numpy array
        '''
        pass

    @abstractmethod
    def label_tile(self, image_nr, pos_zxy, size_zxy, label_value):
        '''
        Get 3d zxy boolean matrix where positions of the requested label
        are indicated with True. Only mapped labelvalues can be requested.

        dimension order: (z, x, y)

        Parameters
        ----------
        image_nr : int
            Index of image.
        pos_zxy : (zslice, x, y)
            Upper left position of subsection.
        size_zxy : (nr_zslices, nr_x, nr_y)
            Upper left position of subsection.
        label_value : int
            Id of the label.

        Returns
        -------
        numpy.ndarray
            3D subsection of labelmatrix as boolean mask in dimension order
            (z, x, y)
        '''

        pass

    @abstractmethod
    def put_tile(self, pixels, pos_zxy, image_nr, label_value):
        '''
        Puts probabilities (pixels) for a certain label to the data storage.

        These probabilities are prediction values from a classifier (i.e. the
        output of the classier)

        Parameters
        ----------
        pixels : numpy.ndarray
            3D matrix of probability values with shape (z, x, y)
        pos_zxy : (z, x, y)
            Upper left position of pixels in source image_nr.
        image_nr : int
            Index of image.
        label_value : int
            Id of the label.

        Returns
        -------
        bool
            True in case of successful write.
        '''

    @abstractmethod
    def image_dimensions(self, image_nr):
        '''
        Get dimensions of the dataset.


        Parameters
        ----------
        image_nr : int
            index of image

        Returns
        -------
        (nr_channels, nr_zslices, nr_x, nr_y)
            Image shape.
        '''
        pass



from yapic_io.tiff_connector import TiffConnector
from pathlib import Path
import numpy as np
import collections
import logging
import os
from itertools import zip_longest
from functools import lru_cache
import sparse
import typing
import h5py


FilePair = collections.namedtuple('FilePair', ['img', 'lbl'])
logger = logging.getLogger(os.path.basename(__file__))


def reconstruct_layer(layer_array: np.array, shape: tuple) -> np.array:
    """Returns a numpy array corresponding to the data reconstruction from a
    sparse version (COO list) of it.

        Parameters
        ----------
        layer_array: Numpy array
            Sparse array version (COO list) of the layer data
        shape: tuple
            shape of the original layer data array

        Returns
        -------
        np.array
            Full version of the layer data
    """
    coords = layer_array[:-1]
    values = layer_array[-1]
    tmp_sparse = sparse.COO(coords=coords, data=values, shape=shape)
    return tmp_sparse.todense()


class NapariConnector(TiffConnector):
    def __init__(self, img_filepath, label_filepath, savepath=None):
        # Dictionary of list telling labeled slices (non-zero matrices)
        self.labeled_slices = dict()
        super().__init__(img_filepath, label_filepath, savepath=savepath)

    def _assemble_filenames(self, pairs):
        self.filenames = [FilePair(Path(img), lbl)
                          for img, lbl in pairs if lbl]
        print('filenames in napariconnector')
        print(self.filenames)

    def _handle_lbl_filenames(self, label_filepath):
        self.h5 = NapariStorage(h5_path=label_filepath, max_dim=4)
        lbl_filenames = self.h5.get_labels_names()
        return label_filepath, lbl_filenames

    def __repr__(self):
        infostring = \
            'NapariConnector object\n' \
            'image filepath: {}\n' \
            'label filepath: {}\n'\
            'nr of images: {}\n'\
            'labelvalue_mapping: {}'.format(self.img_path,
                                            self.label_path,
                                            self.image_count(),
                                            self.labelvalue_mapping)
        return infostring

    def _new_label(self, label_value):

        new_list = []
        new_list = [x for x in label_value[1] if x[1] is not None]

        for x in label_value:
            if label_value[1] is not None:
                new_list.append(x)
            else:
                pass
        label_value = new_list
        return label_value

    def effective_slices(self):
        return self.labeled_slices

    def filter_labeled(self):
        '''
        Removes images without labels.

        Returns
        -------
        NapariConnector
            Connector object containing only images with labels.
        '''
        pairs = [self.filenames[i]for i in range(
            self.image_count()) if self.label_count_for_image(i)]
        tiff_sel = [self.img_path / pair.img for pair in pairs]

        return NapariConnector(tiff_sel, self.label_path,
                               savepath=self.savepath)

    def split(self, fraction, random_seed=42):
        '''
        Split the images pseudo-randomly into two Connector subsets.

        The first of size `(1-fraction)*N_images`, the other of size
        `fraction*N_images`

        Parameters
        ----------
        fraction : float
        random_seed : float, optional

        Returns
        -------
        connector_1, connector_2
        '''

        img_fnames1, img_fnames2, mask = self._split_img_fnames(
            fraction, random_seed=random_seed)

        conn1 = NapariConnector(img_fnames1, self.label_path,
                                savepath=self.savepath)
        conn2 = NapariConnector(img_fnames2, self.label_path,
                                savepath=self.savepath)

        # ensures that both resulting connectors have the same
        # labelvalue mapping (issue #1)
        conn1.labelvalue_mapping = self.labelvalue_mapping
        conn2.labelvalue_mapping = self.labelvalue_mapping

        return conn1, conn2

    def label_tile(self, image_nr, pos_zxy, size_zxy, label_value):
        Z, X, Y = pos_zxy
        ZZ, XX, YY = np.array(pos_zxy) + size_zxy
        _, original_label_value = self._mapped_label_value_to_original(
            label_value)

        slices = self._open_label_file(image_nr)
        if slices is None:
            # return tile with False values
            return np.zeros(size_zxy) != 0
        tile = slices[Z: ZZ, Y: YY, X: XX, 0]
        tile = np.moveaxis(tile, (0, 1, 2), (0, 2, 1))

        tile = (tile == original_label_value)
        return tile

    def _open_label_file(self, image_nr):
        label_filename = self.filenames[image_nr].lbl

        if label_filename is None:
            logger.warning(
                'no label matrix file found for image file %s', str(image_nr))
            return None

        logger.debug('Trying to load labelmat {} in {} Napari project'.format(
            label_filename, self.label_path))

        label_data = self.h5.get_array_data('labels', label_filename)
        label_n_dim = self.h5.n_dims('labels', label_filename)

        if label_n_dim == 2:
            label_data = np.expand_dims(label_data, axis=-1)  # channel
            label_data = np.expand_dims(label_data, axis=0)  # z dim
        elif label_n_dim == 3:
            label_data = np.expand_dims(label_data, axis=-1)  # channel

        if image_nr not in self.labeled_slices.keys():
            self.labeled_slices[image_nr] = self.h5.filled_slices(
                'labels', label_filename)

        return label_data

    def label_matrix_dimensions(self, image_nr):
        '''
        Get dimensions of the label image.


        Parameters
        ----------
        image_nr : int
            index of image

        Returns
        -------
        (nr_channels, nr_zslices, nr_x, nr_y)
            Labelmatrix shape.
        '''
        lbl = self._open_label_file(image_nr)
        if lbl is None:
            return

        output = list(lbl.shape)
        ch = output.pop()
        output[-2], output[-1] = output[-1], output[-2]
        output.insert(0, ch)

        return output

    @lru_cache(maxsize=1)
    def original_label_values_for_all_images(self):
        '''
        Get all unique label values per image.

        Returns
        -------
        list
            List of sets. Each set corresponds to 1 label channel.
            each set contains the label values of that channel.
            E.g. `[{91, 109, 150}, {90, 100}]` for two label channels
        '''
        labels_per_channel = []

        for image_nr in range(self.image_count()):
            label_filename = str(self.filenames[image_nr].lbl)

            if label_filename is None:
                msg = 'No label matrix file found for image file #{}.'
                logger.warning(msg.format(image_nr))
                return None
            print('label filename')
            print(label_filename)
            lbl = self._open_label_file(image_nr)

            lbl = np.transpose(lbl, (3, 0, 2, 1)).astype(int)

            C = lbl.shape[0]
            labels = [np.unique(lbl[c, ...]) for c in range(C)]
            labels = [set(labels) - {0} for labels in labels]

            labels_per_channel = [l1.union(l2)
                                  for l1, l2 in zip_longest(labels_per_channel,
                                                            labels,
                                                            fillvalue=set())]

        return labels_per_channel

    @lru_cache(maxsize=1500)
    def label_count_for_image(self, image_nr):
        '''
        Get number of labels per labelvalue for an image.

        Parameters
        ----------
        image_nr : int
            index of image

        Returns
        -------
        dict
        '''
        label_filename = str(self.filenames[image_nr].lbl)

        if label_filename is None:
            msg = 'No label matrix file found for image file #{}.'
            logger.warning(msg.format(image_nr))
            return None

        lbl = self._open_label_file(image_nr)

        lbl = np.transpose(lbl, (3, 0, 2, 1)).astype(int)

        C = lbl.shape[0]
        labels = [np.unique(lbl[c, ...]) for c in range(C)]

        original_label_count = [{l: np.count_nonzero(lbl[c, ...] == l)
                                 for l in labels[c] if l > 0}
                                for c in range(C)]
        label_count = {self.labelvalue_mapping[c][l]: count
                       for c, orig in enumerate(original_label_count)
                       for l, count in orig.items()}
        return label_count


class NapariStorage():
    def __init__(self, h5_path, max_dim=np.inf):
        self.f = h5py.File(h5_path, 'r')
        self.max_dim = max_dim

    def __iter__(self):
        '''
        Returns (filename, data, type) of the napari layer
        '''
        for layer_type in self.f:  # check folder of Napari layer types
            # only images and labels are required in YAPiC
            if layer_type in ['image', 'labels']:
                # iterate over specific layer
                for layer_name in self.f[layer_type]:
                    if self.dim_check(layer_type, layer_name):
                        layer_data = self.get_array_data(
                            layer_type, layer_name)
                        yield layer_name, layer_data, layer_type

    def get_array_data(self, layer_type: str, layer_name: str) -> np.array:
        '''
        Returns array data of Napari layer. it considers only image and label
        Napari layers (those supported by YAPiC)
            the array dimensions are (?, z, y, x, c)
        '''
        try:
            assert layer_type in ['image', 'labels']
        except AssertionError:
            raise ValueError(
                'Supported Napari layers are image and labels only')
        else:
            try:
                assert layer_name in self.f[layer_type]
            except AssertionError:
                raise ValueError(
                    'There is no Napari layer with the name {}'.format(
                        layer_name))

        napari_layer = self.f[layer_type][layer_name]
        if layer_type == 'labels':
            original_shape = tuple(napari_layer.attrs['shape'])
            if napari_layer.attrs['is_sparse']:
                array_data = reconstruct_layer(np.array(napari_layer),
                                               original_shape)
            else:
                array_data = napari_layer[:]
        else:
            array_data = np.array(napari_layer)
        return array_data
    # This function might be changed in case the hdf5
    # compression protocol is changed! 
    def filled_slices(self, layer_type: str, layer_name: str) -> list:
        '''
        Returns a list of indices specifying which slices have labels
        (values different than 0)
        '''
        napari_layer = self.f[layer_type][layer_name]
        data = np.array(napari_layer)
        if layer_type == 'labels':
            if napari_layer.attrs['is_sparse']:
                # check if the sparse label includes z-dim
                if napari_layer.shape[0] > 3:
                    return list(np.unique(data[0, :]))
            else:
                if len(data.shape) > 2:
                    return [z_slice for z_slice in range(data.shape[0]) 
                            if len(np.unique(data[0, :])) > 1]
        return [0]  # when the images are 2D there is only one slice

    def excluded_layers(self) -> dict:
        '''
        Returns a dictionary of excluded layers due to the maximum dimension
        and not used layer types.

        keys = Napari layer types
        values = Set of skipped layers
        '''
        skipped_types = set(self.f.keys()) - {'image', 'labels'}
        output_dict = {layer_type: set(
            self.f[layer_type].keys()) for layer_type in skipped_types}
        for layer_type in ['image', 'labels']:
            tmp_names = [name for name in self.f[layer_type].keys(
            ) if not self.dim_check(layer_type, name)]
            if len(tmp_names) > 0:
                output_dict[layer_type] = set(tmp_names)
        return output_dict

    def n_dims(self, layer_type: str, layer_name: str) -> int:
        '''
        Returns the number of dimensions of Napari layer
        '''
        napari_layer = self.f[layer_type][layer_name]
        if layer_type == 'labels' and napari_layer.attrs['is_sparse']:
            shape = napari_layer.attrs['shape']
        else:
            shape = napari_layer.shape
        return len(shape)

    def dim_check(self, layer_type: str, layer_name: str) -> bool:
        return bool(self.n_dims(layer_type, layer_name) <= self.max_dim)

    def get_labels_names(self) -> list:
        '''
        Returns a list of all available Napari label layers considering
        the maximum dimension.
        '''
        return [lbl_name for lbl_name in self.f['labels'].keys()
                if self.dim_check('labels', lbl_name)]

    def get_image_names(self) -> list:
        '''
        Returns a list of all available Napari image layers considering
        the maximum dimension.
        '''
        return [im_name for im_name in self.f['image'].keys()
                if self.dim_check('image', im_name)]

    def __len__(self):
        '''
        Returns the number of image and labels (Those supported by YAPiC)
        in the Napari project considering the maximum dimension.
        '''
        return self.number_of_labels() + self.number_of_images()

    def number_of_labels(self) -> int:
        '''
        Returns the number of label layers in the Napari project
        considering the maximum dimension.
        '''
        return len(self.get_labels_names())

    def number_of_images(self) -> int:
        '''
        Returns the number of image layers in the Napari project
        considering the maximum dimension.
        '''
        return len(self.get_image_names())



'''
matrix transformation functions
'''

import numpy as np
from skimage import transform as tf
import logging

logger = logging.getLogger('utils')


def get_transform(image, rotation_angle, shear_angle):
    '''
    Get transformation object for centered rotation and shear of a 2D matrix.

    Paramters
    ---------
    image : numpy.ndarray
        2-dimensional matrix.
    rotation_angle : float
        Angle in degrees.
    shear_angle: float
        angle in degrees

    Returns
    -------
    transformation object
    '''
    if len(image.shape) != 2:
        msg = 'Image has {} dimensions, must have 2.'.format(image.ndim)
        raise ValueError(msg)

    shift_y, shift_x = np.array(image.shape[:2]) / 2.
    tf_rotate_shear = tf.AffineTransform(rotation=np.deg2rad(rotation_angle),
                                         shear=np.deg2rad(shear_angle))
    tf_shift = tf.AffineTransform(translation=[-shift_x, -shift_y])
    tf_shift_inv = tf.AffineTransform(translation=[shift_x, shift_y])
    tf_center_rot = (tf_shift + (tf_rotate_shear + tf_shift_inv)).inverse

    return tf_center_rot


def warp_image_2d(image, rotation_angle, shear_angle):
    '''
    Warps 2d matrix with affine transform.

    Rotation and shear axis is in image center.
    Empty edges are filled by mirroring.

    Paramters
    ---------
    image : numpy.ndarray
        2-dimensional matrix
    rotation_angle: float
        angle in degrees
    shear_angle: float
        angle in degrees

    Returns
    -------
    numpy.ndarray
        transformed 2d matrix
    '''
    if len(image.shape) != 2:
        msg = 'Image has {} dimensions, must have 2.'.format(image.ndim)
        raise ValueError(msg)

    t = get_transform(image, rotation_angle, shear_angle)
    return tf.warp(image, t, order=0, mode='symmetric', preserve_range=True)

def warp_image_2d_stack(image, rotation_angle, shear_angle):
    '''
    Warps a 3d or 4d matrix with affine transform.

    Transformation is applied in 2D slice by sclice.
    Rotation and shear axis is in image center.
    Empty edges are filled by mirroring.

    Paramters
    ---------
    image : numpy.ndarray
        3-dimensional matrix or 4-dimensional matrix
    rotation_angle : float
        angle in degrees
    shear_angle : float
        angle in degrees

    Returns
    -------
    nump.ndarray
        transformed 3d matrix
    '''

    if image.ndim == 3:
        return np.array([warp_image_2d(z_slice, rotation_angle, shear_angle)
                         for z_slice in image])

    elif image.ndim == 4:
        out = []
        for image_zxy in image:
            out.append([warp_image_2d(z_slice, rotation_angle, shear_angle)
                        for z_slice in image_zxy])
        return np.array(out)
    msg = 'Image has {} dimensions, must have 3 or 4.'.format(image.ndim)
    raise ValueError(msg)


def flip_image_2d_stack(image, fliplr=False, flipud=False, rot90=0):
    '''
    Flips and rotates a zxy stack in xy with fast numpy operations.

    Parameters
    ----------
    image : array_like
        must be at least 2D
    fliplr : bool, optional
        If True, image is flipped vertically.
    fliplr : bool, optional
        If True, image is flipped upside down.
    rot90 : integer
        Number of times the array is rotated by 90 degrees.

    Returns
    -------
    numpy.ndarray
        transformed image
    '''
    image = image.T
    if fliplr:
        image = np.fliplr(image)
    if flipud:
        image = np.flipud(image)
    if rot90 > 0:
        image = np.rot90(image, k=rot90)
    return image.T



import yapic_io.utils as ut
import numpy as np
from numpy.testing import assert_equal
import os
import logging
from yapic_io.minibatch import Minibatch

logger = logging.getLogger(os.path.basename(__file__))


class PredictionBatch(Minibatch):
    '''
    List-like data interface for classification with neural networks.

    * Provides ``get_pixels()`` method for getting pixel tiles from
      4D images ``(channels, z, x, y)`` that fit into your neural network
      input layer.
    * Pixel data is loaded lazily to support arbitrary large image datasets.
    * Provides ``put_probmap_data()`` method to transfer classification
      results of your neural network back to the data source.
    * Flexible data binding through the Dataset object.

    Parameters
    ----------
    dataset: Dataset
        Handle for reading and writing pixels.
    batch_size: int
        Nr of tiles for one batch.
    size_zxy: (nr_zslices, nr_x, nr_y)
        3d tile size (size of classifier output template).
    padding_zxy: (z, x, y)
        Growing of pixel tile in z, x and y (to define size of classifier input
        template relative to output template size size_zxy).

    Examples
    --------
    >>> from yapic_io import TiffConnector, Dataset, PredictionBatch
    >>> import tempfile
    >>> # mock classification function
    >>> def classify(pixels, value):
    ...     return np.ones(pixels.shape) * value
    >>>
    >>>
    >>> # define data loacations
    >>> pixel_img_dir = 'yapic_io/test_data/tiffconnector_1/im/*.tif'
    >>> label_img_dir = 'yapic_io/test_data/tiffconnector_1/labels/*.tif'
    >>> savepath = tempfile.TemporaryDirectory()
    >>>
    >>> # size of network output layer in zxy
    >>> tile_size = (1, 5, 4) # size of network output layer in zxy
    >>>
    >>> # padding of network input layer in zxy, in respect to output layer
    >>> padding = (0, 0, 0)
    >>>
    >>> # make training_batch mb and prediction interface p
    >>> # with TiffConnector binding
    >>> c = TiffConnector(pixel_img_dir, label_img_dir, savepath=savepath.name)
    >>> p = PredictionBatch(Dataset(c), 2, tile_size, padding_zxy=padding)
    >>> len(p)
    255
    >>> p.labels
    {1, 2, 3}
    >>>
    >>> # classify the whole bound dataset
    >>> counter = 0 # needed for mock data
    >>> for item in p:
    ...     pixels = item.pixels() # input for classifier
    ...     mock_classifier_result = classify(pixels, counter)
    ...     # pass classifier results for each class to data source
    ...     item.put_probmap_data(mock_classifier_result)
    ...     counter += 1

    See Also
    --------
    yapic.io.training_batch
    '''
    def __init__(self, dataset, batch_size, size_zxy, padding_zxy=(0, 0, 0)):

        super().__init__(dataset,
                         batch_size,
                         size_zxy,
                         padding_zxy=padding_zxy)
        self.current_batch_pos = 0
        self.multichannel = False

        if size_zxy:
            self.set_tile_size(size_zxy)

    def multichannel_output_on(self):
        '''
        Probability maps are saved in multichannel images, one channel for each
        class.
        '''
        self.multichannel = True

    def multichannel_output_off(self):
        '''
        Probability maps are saved separate singlechannel images, one image for
        each class with suffix _class{class_nr}.tif.
        '''
        self.multichannel = False

    def set_tile_size(self, size_zxy):
        super().set_tile_size(size_zxy)
        self._all_tile_positions = self._compute_pos_zxy()

    def pixels(self):
        load_img = self.dataset.multichannel_pixel_tile

        pixels = [load_img(im_nr, pos_zxy,
                           self.tile_size_zxy,
                           self.channels,
                           self.padding_zxy)
                  for im_nr, pos_zxy in self.current_tile_positions]
        pixels = np.array(pixels)
        pixels = np.moveaxis(pixels, [0, 1, 2, 3, 4],
                             self.pixel_dimension_order)

        return self._normalize(pixels).astype(self.float_data_type)

    def __len__(self):
        '''
        Return the number of batches.
        '''
        n = len(self._all_tile_positions)  # nr of single tiles
        return int(np.ceil(n / self._batch_size))

    def __getitem__(self, position):
        '''
        Implements list-like operations of element selection and slicing.
        '''
        if position >= len(self):
            raise IndexError('index out of bounds')

        self.current_batch_pos = position
        return self

    @property
    def current_tile_positions(self):
        total = len(self._all_tile_positions)
        start = self.current_batch_pos * self._batch_size
        size = np.minimum(self._batch_size, total - start)

        return [self._all_tile_positions[x]
                for x in range(start, start + size)]

    def put_probmap_data(self, probmap_batch):
        '''
        Put classification results to the data source.



        Parameters
        ----------
        probmap_batch: ndarray
            5D matrix with shape (batches, nr_labels, z, x, y).

        Returns
        -------
        bool
            True if successful.

        Notes
        -----
        The order of the labels list (acessed with ``self.labels``) defines
        the order of the labels layer in the probability map.

        To pass 3D probmaps for a certain label, use
        ``put_probmap_data_for_label()``.
        '''

        # reorder dimensions to bczxy
        probmap_batch = np.moveaxis(probmap_batch,
                                    self.pixel_dimension_order,
                                    [0, 1, 2, 3, 4])
        nr_classes = probmap_batch.shape[1]
        if not self.multichannel:
            nr_classes = False

        assert_equal(len(probmap_batch.shape), 5, '5-dim (B,L,Z,X,Y) expected')
        B, L, *ZXY = probmap_batch.shape
        self.labels = np.arange(L) + 1 if len(self.labels) == 0 \
            else self.labels

        assert_equal(B, len(self.current_tile_positions))
        assert_equal(L, len(self.labels))
        assert_equal(ZXY, self.tile_size_zxy)

        for probmap, (image_nr, pos_zxy) in zip(probmap_batch,
                                                self.current_tile_positions):

            for label_ch, label in zip(probmap, self.labels):
                self.dataset.pixel_connector.put_tile(
                    label_ch,
                    pos_zxy,
                    image_nr,
                    label,
                    multichannel=nr_classes)

    def _compute_pos_zxy(self):
        '''
        Compute all possible tile positions for the whole dataset
        for tile_size = self._size_zxy (size of output layer).

        Returns
        -------
        list
            Tile positions
            e.g.
            ``[(image_nr, zpos, xpos, ypos),
               (image_nr, zpos, xpos, ypos), ...]``
        '''
        tile_pos = []
        for img_nr in list(range(self.dataset.n_images)):
            img_shape_zxy = self.dataset.image_dimensions(img_nr)[1:]
            tile_pos = tile_pos + [(img_nr, pos) for pos
                                   in ut.compute_pos(img_shape_zxy,
                                                     self.tile_size_zxy)]

        return tile_pos



import logging
import os
import collections
from functools import lru_cache
import yapic_io.utils as ut
import numpy as np
import itertools
import warnings
from itertools import zip_longest
from pathlib import Path
from yapic_io.connector import Connector

from tifffile import memmap, TiffFile

logger = logging.getLogger(os.path.basename(__file__))

FilePair = collections.namedtuple('FilePair', ['img', 'lbl'])


def _handle_img_filenames(img_filepath):
    '''
    - checks if list of image filepaths, a single wildcard filepath
      or a single filepath without a wildcard is given.
    - checks if given filenames exit
    - splits into folder and list of filenames
    '''

    if type(img_filepath) in (str, Path):
        img_filepath = os.path.abspath(img_filepath)
        img_filemask = '*.tif' if os.path.isdir(img_filepath)\
            else os.path.basename(img_filepath)

        folder = Path(img_filepath if os.path.isdir(img_filepath)
                      else os.path.dirname(img_filepath))
        filenames = [fname.name for fname in sorted(folder.glob(img_filemask))]

    elif type(img_filepath) in (list, tuple):

        img_filenames = img_filepath
        img_filenames = [Path(p).expanduser().resolve()
                         if p is not None else None
                         for p in img_filepath]

        assert len(img_filenames) > 0, 'list of image filenames is empty'

        for e in img_filenames:
            if e is not None:
                assert e.exists(), 'file {} not found'.format(e)

        folders = {fname.parent
                   for fname in img_filenames if fname is not None}
        assert len(folders) == 1, 'image filenames are not in the same folder'
        folder = next(iter(folders))
        folder = folder.expanduser().resolve()
        filenames = [fname.name
                     if fname is not None else None
                     for fname in img_filenames]

    else:
        raise NotImplementedError(
            'could not import images from {}'.format(img_filepath))

    logger.info('{} image files detected.'.format(len(filenames)))
    return folder, filenames


class TiffConnector(Connector):
    '''
    Implementation of Connector for tiff images up to 4 dimensions and
    corresponding label masks up to 4 dimensions in tiff format.

    Parameters
    ----------
    img_filepath : str or list of str
        Path to source pixel images (use wildcards for filtering)
        or a list of filenames.
    label_filepath : str or list of str
        Path to label images (use wildcards for filtering)
        or a list of filenames.
    savepath : str, optional
        Directory to save pixel classifiaction results as probability
        images.

    Notes
    -----
    Label images and pixel images have to be equal in zxy dimensions,
    but can differ in nr of channels.

    Labels can be read from multichannel images. This is needed for
    networks with multiple output layers. Each channel is assigned one
    output layer. Different labels from different channels can overlap
    (can share identical xyz positions).

    Examples
    --------
    Create a TiffConnector object with pixel and label data.

    >>> from yapic_io.tiff_connector import TiffConnector
    >>> pixel_image_dir = 'yapic_io/test_data/tiffconnector_1/im/*.tif'
    >>> label_image_dir = 'yapic_io/test_data/tiffconnector_1/labels/*.tif'
    >>> t = TiffConnector(pixel_image_dir, label_image_dir)
    >>> print(t)
    TiffConnector object
    image filepath: yapic_io/test_data/tiffconnector_1/im
    label filepath: yapic_io/test_data/tiffconnector_1/labels
    nr of images: 3
    labelvalue_mapping: [{91: 1, 109: 2, 150: 3}]

    See Also
    --------
    yapic_io.ilastik_connector.IlastikConnector
    '''

    def __init__(self, img_filepath, label_filepath, savepath=None):

        self.img_path, img_filenames = _handle_img_filenames(img_filepath)
        self.label_path, lbl_filenames = self._handle_lbl_filenames(
            label_filepath)

        assert img_filenames is not None, 'no filenames for pixel images found'
        assert len(img_filenames) != 0, 'no filenames for pixel images found'

        if lbl_filenames is None or len(lbl_filenames) == 0:
            pairs = [(img, None) for img in img_filenames]
        else:
            pairs = ut.find_best_matching_pairs(img_filenames, lbl_filenames)

        self._assemble_filenames(pairs)

        logger.info('Pixel and label files are assigned as follows:')
        logger.info('\n'.join('{p.img} <-> {p.lbl}'.format(p=pair)
                              for pair in self.filenames))

        self.savepath = Path(savepath) if savepath is not None else None

        original_labels = self.original_label_values_for_all_images()
        self.labelvalue_mapping = self.calc_label_values_mapping(
            original_labels)

        self.check_label_matrix_dimensions()

    def _assemble_filenames(self, pairs):
        self.filenames = [FilePair(Path(img), Path(lbl) if lbl else None)
                          for img, lbl in pairs]

    def _handle_lbl_filenames(self, label_filepath):
        return _handle_img_filenames(label_filepath)

    def __repr__(self):

        infostring = \
            'TiffConnector object\n' \
            'image filepath: {}\n' \
            'label filepath: {}\n'\
            'nr of images: {}\n'\
            'labelvalue_mapping: {}'.format(self.img_path,
                                            self.label_path,
                                            self.image_count(),
                                            self.labelvalue_mapping)
        return infostring

    def filter_labeled(self):
        '''
        Removes images without labels.

        Returns
        -------
        TiffConnector
            Connector object containing only images with labels.
        '''
        img_fnames = [self.img_path / img for img, lbl in self.filenames
                      if lbl is not None]

        lbl_fnames = [self.label_path / lbl
                      for img, lbl in self.filenames
                      if lbl is not None]

        return TiffConnector(img_fnames, lbl_fnames, savepath=self.savepath)

    def _split_img_fnames(self, fraction, random_seed=42):
        # i took this out from the split method to be used in split method
        # of child methods (e.g. IlasikConnector)
        N = len(self.filenames)

        state = np.random.get_state()
        np.random.seed(random_seed)
        mask = np.random.choice([True, False], size=N, p=[
                                1 - fraction, fraction])
        np.random.set_state(state)

        img_fnames1 = [self.img_path / img
                       for img, lbl in itertools.compress(self.filenames,
                                                          mask)]

        img_fnames2 = [self.img_path / img
                       for img, lbl in itertools.compress(self.filenames,
                                                          ~mask)]

        if len(img_fnames1) == 0:
            msg = ('TiffConnector.split({}): ' +
                   'First connector is empty!').format(fraction)
            warnings.warn(msg)

        if len(img_fnames2) == 0:
            msg = ('TiffConnector.split({}): ' +
                   'Second connector is empty!').format(fraction)
            warnings.warn(msg)

        return img_fnames1, img_fnames2, mask

    def split(self, fraction, random_seed=42):
        '''
        Split the images pseudo-randomly into two Connector subsets.

        The first of size `(1-fraction)*N_images`, the other of size
        `fraction*N_images`

        Parameters
        ----------
        fraction : float
        random_seed : float, optional

        Returns
        -------
        connector_1, connector_2
        '''

        img_fnames1, img_fnames2, mask = self._split_img_fnames(
            fraction,
            random_seed=random_seed)

        lbl_fnames1 = [self.label_path / lbl if lbl is not None else None
                       for img, lbl in itertools.compress(self.filenames,
                                                          mask)]
        lbl_fnames2 = [self.label_path / lbl if lbl is not None else None
                       for img, lbl in itertools.compress(self.filenames,
                                                          ~mask)]

        conn1 = TiffConnector(img_fnames1, lbl_fnames1, savepath=self.savepath)
        conn2 = TiffConnector(img_fnames2, lbl_fnames2, savepath=self.savepath)

        # ensures that both resulting tiff_connectors have the same
        # labelvalue mapping (issue #1)
        conn1.labelvalue_mapping = self.labelvalue_mapping
        conn2.labelvalue_mapping = self.labelvalue_mapping

        # np.random.seed(None)
        return conn1, conn2

    def image_count(self):
        return len(self.filenames)

    @lru_cache(maxsize=10)
    def _open_probability_map_file(self,
                                   image_nr,
                                   label_value,
                                   multichannel=False):
        # memmap is slow, so we must cache it to be fast!
        fname = self.filenames[image_nr].img
        if multichannel:
            fname = Path('{}.tif'.format(fname.stem))
            n_classes = multichannel
            C = n_classes
        else:
            fname = Path('{}_class_{}.tif'.format(fname.stem, label_value))
            C = 1  # channel in output probmap

        path = self.savepath / fname

        if not path.exists():  # created "empty" tif of shape
            _, Z, X, Y = self.image_dimensions(image_nr)
            return memmap(path, shape=(Z, Y, X, C), dtype='float32')

        return memmap(path)

    def put_tile(self,
                 pixels,
                 pos_zxy,
                 image_nr,
                 label_value,
                 multichannel=False):

        assert self.savepath is not None
        np.testing.assert_equal(len(pos_zxy), 3)
        np.testing.assert_equal(len(pixels.shape), 3)
        pixels = np.array(pixels, dtype=np.float32)

        slices = self._open_probability_map_file(image_nr,
                                                 label_value,
                                                 multichannel=multichannel)

        C = 0
        if multichannel:
            C = label_value - 1
        Z, X, Y = pos_zxy
        ZZ, XX, YY = np.array(pos_zxy) + pixels.shape

        # Temporal dimension shift in pixels
        # current is z, x, y but we use z, y, x
        pixels = np.moveaxis(pixels, (0, 1, 2), (0, 2, 1))

        slices[Z: ZZ, Y: YY, X: XX, C] = pixels

    @staticmethod
    def fix_dims(memmap_array, path):
        """This method will fix the dimensions of memmap_array
        in the order (Z, Y, X, C).
        It will increment the dimensions if memmap_array has less than
        4 dimensions.
        It uses the tiff file metadata to guess the original dimensions
        order and which dimensions are missing.

        Parameters
        ----------
        memmap_array - numpy.memmap: sorted image data array in the dimension
        order (Z, Y, X, C)
        path - string: Tiff file path

        Returns
        -------
        connector_1, connector_2
        """
        # target dims (Z,Y,X,C)
        with TiffFile(path) as tif:
            axes = tif.series[0].axes

        # Adding the missed axis
        dims_dict = {'T': 'Z', 'S': 'C', 'Q': 'C'}
        # The letter to represent each dimension may change depending on the
        # file generation.
        # We will transform this representation to Z, Y, X, C to generalize
        # the process
        axes = axes.translate(axes.maketrans(dims_dict))
        if 'C' not in axes:
            memmap_array = np.expand_dims(memmap_array, axis=-1)
            axes += 'C'
        if 'Z' not in axes:
            memmap_array = np.expand_dims(memmap_array, axis=0)
            axes = 'Z' + axes

        # Sorting the axes
        dim_map = ['ZYXC'.index(dim) for dim in axes]
        memmap_array = np.moveaxis(memmap_array, (0, 1, 2, 3), dim_map)
        return memmap_array

    @lru_cache(maxsize=10)
    def _open_image_file(self, image_nr):
        """Returns memmap object with shape: z, y, x, c"""
        # memmap is slow, so we must cache it to be fast!
        path = self.img_path / self.filenames[image_nr].img
        im_data = memmap(path)
        return self.fix_dims(im_data, path)  # shape order: z, y, x, c

    def image_dimensions(self, image_nr):
        """returns a tuple representing the size of the image in the
        order of: C, Z, X, Y"""
        img = self._open_image_file(image_nr)
        Z, Y, X, C = img.shape
        return (C, Z, X, Y)

    def label_matrix_dimensions(self, image_nr):
        '''
        Get dimensions of the label image.


        Parameters
        ----------
        image_nr : int
            index of image

        Returns
        -------
        (nr_channels, nr_zslices, nr_x, nr_y)
            Labelmatrix shape.
        '''
        lbl = self._open_label_file(image_nr)
        if lbl is None:
            return

        Z, Y, X, C = lbl.shape
        return (C, Z, X, Y)

    def check_label_matrix_dimensions(self):
        '''
        Check if label matrix dimensions fit to image dimensions, i.e.
        everything identical except nr of channels (label mat always 1).

        Raises
        ------
        AssertionError
            If label matrix dimensions don't fit to image dimensions.
        '''
        N_channels = None

        for i, (img_fname, lbl_fname) in enumerate(self.filenames):
            img_dim = self.image_dimensions(i)
            lbl_dim = self.label_matrix_dimensions(i)

            msg = 'Dimensions for image #{}: img.shape={}, lbl.shape={}'
            logger.debug(msg.format(i, img_dim, lbl_dim))

            if lbl_dim is None:
                continue

            _,  *img_dim = img_dim
            ch, *lbl_dim = lbl_dim

            if N_channels is None:
                N_channels = ch

            msg = 'Label channels inconsistent for {}'.format(lbl_fname)
            np.testing.assert_equal(N_channels, ch, msg)
            msg = 'Invalid image dims for {} and {}'.format(img_fname,
                                                            lbl_fname)
            np.testing.assert_array_equal(lbl_dim, img_dim, msg)

    def _mapped_label_value_to_original(self, label_value):
        '''
        self.labelvalue_mapping in reverse
        '''
        for c, mapping in enumerate(self.labelvalue_mapping):
            reverse_mapping = {v: k for k, v in mapping.items()}
            original = reverse_mapping.get(label_value)
            if original is not None:
                return c, original

        msg = 'Should not be reached! (mapped_label_value={}, mapping={})'
        raise Exception(msg.format(label_value, self.labelvalue_mapping))

    def get_tile(self, image_nr, pos, size):
        """Returns a tile of image as a numpy array.
        the output shape corresponds to: c, z, x, y"""
        C, Z, X, Y = pos
        CC, ZZ, XX, YY = np.array(pos) + size

        slices = self._open_image_file(image_nr)

        tile = np.array(slices[Z: ZZ, Y: YY, X: XX, C: CC])
        tile = np.moveaxis(tile, (0, 1, 2, 3), (1, 3, 2, 0))

        return tile.astype('float')

    def label_tile(self, image_nr, pos_zxy, size_zxy, label_value):
        Z, X, Y = pos_zxy
        ZZ, XX, YY = np.array(pos_zxy) + size_zxy
        C, original_label_value = self._mapped_label_value_to_original(
            label_value)

        slices = self._open_label_file(image_nr)
        if slices is None:
            # return tile with False values
            return np.zeros(size_zxy) != 0

        tile = np.array(slices[Z: ZZ, Y: YY, X: XX, C])
        tile = np.moveaxis(tile, (0, 1, 2), (0, 2, 1))
        tile = (tile == original_label_value)
        return tile

    @lru_cache(maxsize=10)
    def _open_label_file(self, image_nr):
        # memmap is slow, so we must cache it to be fast!
        path = self.img_path / self.filenames[image_nr].img
        label_filename = self.filenames[image_nr].lbl

        if label_filename is None:
            logger.warning(
                'no label matrix file found for image file %s', str(image_nr))
            return None

        path = self.label_path / label_filename
        logger.debug('Trying to load labelmat %s', path)

        lbl_data = memmap(path)

        return self.fix_dims(lbl_data, path)  # shape order: z, y, x, c

    @staticmethod
    def calc_label_values_mapping(original_labels):
        '''
        Assign unique labelvalues to original labelvalues.

        For multichannel label images it might happen, that identical
        labels occur in different channels.
        to avoid conflicts, original labelvalues are mapped to unique values
        in ascending order 1, 2, 3, 4...
        This is defined in self.labelvalue_mapping:

        [{orig_label1: 1, orig_label2: 2}, {orig_label1: 3, orig_label2: 4},..]

        Each element of the list correponds to one label channel.
        Keys are the original labels, values are the assigned labels that
        will be seen by the Dataset object.

        Parameters
        ----------
        original_labels : array_like
            List of original label values.

        Returns
        -------
        dict
            Labelvalue mapping with original labels as key and new label as
            value.
        '''
        new_labels = itertools.count(1)

        label_mappings = [
            {l: next(new_labels) for l in sorted(labels_per_channel)}
            for labels_per_channel in original_labels
        ]

        logger.debug('Label values are mapped to ascending values:')
        logger.debug(label_mappings)
        return label_mappings

    @lru_cache(maxsize=1)
    def original_label_values_for_all_images(self):
        '''
        Get all unique label values per image.

        Returns
        -------
        list
            List of sets. Each set corresponds to 1 label channel.
            each set contains the label values of that channel.
            E.g. `[{91, 109, 150}, {90, 100}]` for two label channels
        '''
        labels_per_channel = []

        for image_nr in range(self.image_count()):
            slices = self._open_label_file(image_nr)
            if slices is None:
                continue

            C = slices.shape[-1]
            labels = [np.unique(slices[..., c]) for c in range(C)]

            labels = [set(labels) - {0} for labels in labels]

            labels_per_channel = [l1.union(l2)
                                  for l1, l2 in zip_longest(labels_per_channel,
                                                            labels,
                                                            fillvalue=set())]

        return labels_per_channel

    @lru_cache(maxsize=1500)
    def label_count_for_image(self, image_nr):
        '''
        Get number of labels per labelvalue for an image.

        Parameters
        ----------
        image_nr : int
            index of image

        Returns
        -------
        dict
        '''
        slices = self._open_label_file(image_nr)
        if slices is None:
            return None

        C = slices.shape[-1]
        labels = [np.unique(slices[..., c]) for c in range(C)]

        original_label_count = [{l: np.count_nonzero(slices[..., c] == l)
                                 for l in labels[c] if l > 0}
                                for c in range(C)]

        label_count = {self.labelvalue_mapping[c][l]: count
                       for c, orig in enumerate(original_label_count)
                       for l, count in orig.items()}
        return label_count



import os
import logging
import numpy as np
from collections.abc import Iterable

logger = logging.getLogger(os.path.basename(__file__))


class Minibatch(object):
    '''
    Iterator providing pixel and label data for classifier training
    or prediction.
    This class is not directly used, but provides common functions
    as a parent class for TrainingBatch and PredictionBatch.

    Parameters
    ----------
    dataset: Dataset
        Handle for pixels and labels weights.
    batch_size: int
        Nr of tiles for one batch.
    size_zxy: (nr_zslices, nr_x, nr_y)
        3d tile size (size of classifier output template).
    padding_zxy: (z, x, y)
        Growing of pixel tile in z, x and y (to define size of classifier input
        template relative to output template size size_zxy).

    See Also
    --------
    yapic.io.training_batch
    yapic_io.prediction_batch
    '''
    def __init__(self, dataset, batch_size, size_zxy, padding_zxy=(0, 0, 0)):

        self.dataset = dataset

        self.pixel_dimension_order = [0, 1, 2, 3, 4]
        self.set_pixel_dimension_order('bczxy')

        self._batch_size = batch_size
        self.normalize_mode = None
        self.global_norm_minmax = None
        self.float_data_type = np.float32  # type of pixel and weight data

        if size_zxy:
            np.testing.assert_equal(len(size_zxy), 3,
                                    'len of size_zxy be 3: (z, x, y)')
        if padding_zxy:
            np.testing.assert_equal(len(padding_zxy), 3,
                                    'len of padding_shoule be 3: (z, x, y)')

        self.tile_size_zxy = size_zxy
        self.padding_zxy = padding_zxy

        # imports all available channels by default
        nr_channels = self.dataset.image_dimensions(0)[0]
        self.channels = set(range(nr_channels))
        # imports all available labels by default
        self.labels = set(self.dataset.label_values())

    def set_pixel_dimension_order(self, s):
        '''
        Parameters
        ----------
        s : {'bczxy', 'bzxyc'}
            Dimension order of pixel array as returned by self.pixels()
            and self.weights().
            First dimension is always the batch dimension and can not be
            changed.

        Notes
        -----
        Keras supports pixel dimension order 'bzxyc'.
        '''

        assert s in ['bczxy',
                     'bzxyc'], 'dimension order {} not supported'.format(s)

        self.pixel_dimension_order[0] = s.find('b')
        self.pixel_dimension_order[1] = s.find('c')
        self.pixel_dimension_order[2] = s.find('z')
        self.pixel_dimension_order[3] = s.find('x')
        self.pixel_dimension_order[4] = s.find('y')

        return self.pixel_dimension_order

    def set_tile_size(self, size_zxy):
        self.tile_size_zxy = size_zxy

    def set_normalize_mode(self, mode_str, minmax=None):
        '''
        Parameters
        ----------
        mode_str: {'local', 'local_z_score', 'global', 'minmax'}
            Normalization mode for pixel values. Channels are treated
            separately.

            * 'local': scale between 0 and 1 per minibatch.
            * 'local_z_score' : mean to zero, scale by standard deviation,
              calculated per minibatch
            * 'global': scale ``minmax=(min, max)`` between 0 and 1,
              calculated for whole dataset
        minmax : (min, max)
            Tuple of minimum and maximum pixel value for global normalization,
            e.g ``(0, 255)`` for 8 bit images, or list of tuples, one for each
            channel, e.g. ``[(0, 255), (0, 100), (0, 255)]``
        '''
        valid = ('local_z_score', 'local', 'off', 'global')
        msg = 'Invalid normalization mode: "{}". Valid choices: {}'
        assert mode_str in valid, msg.format(mode_str, valid)

        self.normalize_mode = mode_str

        if mode_str == 'global':

            if minmax is None:
                # calculate upper and lower percentile automatically
                minmax = self.dataset.pixel_statistics(self.channels)
            n_channels = self.dataset.pixel_connector.image_dimensions(0)[0]

            if len(minmax) == 2:
                if n_channels == 2:
                    assert (_is_twotuple_of_numerics(minmax)
                            or _is_list_of_twotuples(minmax))
                else:
                    assert _is_twotuple_of_numerics(minmax)
            else:
                assert n_channels == len(minmax)
                assert _is_list_of_twotuples(minmax)

            msg = 'Scale reference for pixel normalization must not be zero.'
            if _is_twotuple_of_numerics(minmax):
                assert minmax[1] != 0, msg
            if _is_list_of_twotuples(minmax):
                for e in minmax:
                    assert e[1] != 0, msg

            self.global_norm_minmax = minmax

    def _normalize(self, pixels):
        '''
        To be called by the ``self.pixels()`` function
        '''
        if self.normalize_mode in ('off', None):
            return pixels

        if self.normalize_mode == 'global':
            if _is_twotuple_of_numerics(self.global_norm_minmax):
                # same reference for all channels
                center_ref = np.array(self.global_norm_minmax[0])
                scale_ref = np.array(self.global_norm_minmax[1])
            else:
                center_ref = np.array([e[0] for e in self.global_norm_minmax])
                scale_ref = np.array([e[1] for e in self.global_norm_minmax])

        elif self.normalize_mode == 'local':
            # (data - minval) / (maxval - minval)
            # percentiles are used to be robust against outliers
            min_ref, max_ref = np.percentile(pixels,
                                             [1, 99],
                                             axis=(0, 2, 3, 4))

            center_ref = min_ref
            scale_ref = max_ref - min_ref

        elif self.normalize_mode == 'local_z_score':
            # (data - mean) / std
            center_ref = pixels.mean(axis=(0, 2, 3, 4))
            scale_ref = pixels.std(axis=(0, 2, 3, 4))

        pixels_centered = (pixels.swapaxes(1, 4) - center_ref)

        if (scale_ref == 0).all():
            # if scale_ref is zero, do not scale to avoid zero division
            return pixels_centered.swapaxes(1, 4)
        else:
            # scale by scale reference
            return (pixels_centered / scale_ref).swapaxes(1, 4)


def _is_twotuple_of_numerics(tp):
    if not isinstance(tp, Iterable):
        return False
    if not len(tp) == 2:
        return False
    for e in tp:
        if not (isinstance(e, int) or isinstance(e, float)):
            return False
    return True


def _is_list_of_twotuples(ls):
    if not isinstance(ls, Iterable):
        return False
    for e in ls:
        if not _is_twotuple_of_numerics(e):
            return False
    return True



from abc import ABCMeta, abstractmethod
from yapic_io.connector import Connector


class CoordinateConnector(Connector, metaclass=ABCMeta):
    '''
    Interface to pixel and label data source for classifier
    training and prediction.
    The connector is optimized for data sources where labels are stored
    in coordinates.

    The Coordinate_connector class has the 'by_label_index' mode to
    fetch training tiles. In this mode a dataset object fetches pixel
    and label tiles by a specific label coordinate.
    This is potentially more efficient than random polling in the following
    situation:

    - very large pixel images (e.g. slide scanner data)
    - sparse labelling where labels are stored as coordinates
    '''

    @abstractmethod
    def label_index_to_coordinate(self, image_nr, label_value, label_index):
        '''
        Get image coordinate for specific label.

        Parameters
        ----------
        image_nr : int
            Index of image.
        label_value : int
            Id of the label.
        label_index: int
            Value between 0 and count[label_value].
            Label count can be retrieved with self.label_count_for_image
            method.

        Returns
        -------
        ndarray
            czxy coordinate of a specific label (specified by the
            label index) with labelvalue label_value (mapped label value).
        '''

        pass



import random
from yapic_io.napari_connector import NapariConnector
import numpy as np
from yapic_io.minibatch import Minibatch
from yapic_io.utils import compute_pos, find_overlapping_tiles, progressbar
import logging
import os

logger = logging.getLogger(os.path.basename(__file__))
logger.setLevel(logging.INFO)


class TrainingBatch(Minibatch):
    '''
    Infinite iterator providing pixel and label data for classifier training.

    * Provides tile data for classifier training.
    * Data is loaded from the dataset object.

    Parameters
    ----------
    dataset: Dataset
        Handle for reading pixels and labels weights.
    batch_size: int
        Nr of tiles for one batch.
    size_zxy: (nr_zslices, nr_x, nr_y)
        3d tile size (size of classifier output template).
    padding_zxy: (z, x, y)
        Growing of pixel tile in z, x and y (to define size of classifier input
        template relative to output template size size_zxy).
    equalized: bool
        If ``True``, less frequent labels are favored in randomized
        tile selection.

    Examples
    --------
    >>> from yapic_io import TiffConnector, Dataset, PredictionBatch
    >>> import tempfile
    >>>
    >>> # define data locations
    >>> pixel_img_dir = 'yapic_io/test_data/tiffconnector_1/im/*.tif'
    >>> label_img_dir = 'yapic_io/test_data/tiffconnector_1/labels/*.tif'
    >>> savepath = tempfile.TemporaryDirectory()
    >>>
    >>> # size of network output layer in zxy
    >>> tile_size = (1, 5, 4) # size of network output layer in zxy
    >>>
    >>> # padding of network input layer in zxy, in respect to output layer
    >>> padding = (0, 2, 2)
    >>>
    >>> c = TiffConnector(pixel_img_dir, label_img_dir, savepath=savepath.name)
    >>> m = TrainingBatch(Dataset(c), tile_size, padding_zxy=padding)
    >>> print(m)
    TrainingBatch (batch_size: 3, tile_size (zxy): (1, 5, 4), augment: {'flip'}
    >>>
    >>> for counter, mini in enumerate(m):
    ...     # shape of weights is (6, 3, 1, 5, 4) :
    ...     # batchsize 6 , 3 label-classes, 1 z, 5 x, 4 y
    ...     weights = mini.weights()
    ...
    ...     pixels = mini.pixels()
    ...     # shape of pixels is (6, 3, 1, 9, 8) :
    ...     # 3 channels, 1 z, 9 x, 4 y (more xy due to padding)
    ...
    ...     # training on mini.pixels and mini.weights goes here
    ...     if counter > 10: #m is infinite
    ...         break
    '''

    def __init__(self,
                 dataset,
                 size_zxy,
                 padding_zxy=(0, 0, 0),
                 equalized=False):

        batch_size = len(dataset.label_values())

        super().__init__(dataset,
                         batch_size,
                         size_zxy,
                         padding_zxy=padding_zxy)

        self.equalized = equalized
        self.augmentation = set()
        self.augment_by_flipping(True)
        self.rotation_range = None
        self.shear_range = None
        self._pixels = None
        self._weights = None

        self.tile_pos_for_label = {key: self.tile_positions(sliding=True)
                                   for key in self.labels}

    def __repr__(self):
        info = ('TrainingBatch (batch_size: {}, '
                'tile_size (zxy): {}, augment: {}')
        return info.format(self._batch_size,
                           self.tile_size_zxy,
                           self.augmentation)

    def __iter__(self):
        return self

    def __next__(self):
        pixels = []
        weights = []
        augmentations = []

        for label in self.labels:
            tile_data = self._random_tile(for_label=label)

            pixels.append(tile_data.pixels)
            weights.append(tile_data.weights)
            augmentations.append(tile_data.augmentation)

        self._pixels = np.array(pixels)
        self._weights = np.array(weights, self.float_data_type)
        self.augmentations = augmentations

        return self

    def effective_tiles(self):
        """This function discards those tiles selected from slices that
        were not labeled. It must be used only when the pixel_connector
        is NapariConnector."""
        labeled_slices = self.dataset.pixel_connector.effective_slices()
        # list of tuples with (img_id, z_slice)
        slices_with_ids = []
        for img_id, value in labeled_slices.items():
            tmp = [(img_id, v) for v in value]
            slices_with_ids += tmp
        output = dict()
        for label_id, tiles in self.tile_pos_for_label.items():
            output[label_id] = [
                tile for tile in tiles if tile[:2] in slices_with_ids]
        return output

    def augment_by_flipping(self, flip_on):
        '''
        Data augmentation setting. Advantage: fast.

        Parameters
        ----------
        flip_on: bool
            If ``True``, tiles are randomly flipped for augmentation.
        '''
        if flip_on:
            self.augmentation.add('flip')
        else:
            self.augmentation.discard('flip')

    def augment_by_rotation(self, rot_on, rotation_range=(-45, 45)):
        '''
        Data augmentation setting. Slower than flipping, but you get more
        training data.

        Parameters
        ----------
        rot_on: bool
            If ``True``, tiles are randomly rotated.
        rotation_range: (min, max)
            Rotation angle range in degrees.
        '''
        self.rotation_range = rotation_range
        if rot_on:
            self.augmentation.add('rotate')
        else:
            self.augmentation.discard('rotate')

    def augment_by_shear(self, shear_on, shear_range=(-5, 5)):
        '''
        Data augmentation setting. Slower than flipping, but you get more
        training data.

        Parameters
        ----------
        shear_on: bool
            If ``True``, tiles are randomly sheared.
        shear_range: (min, max)
            Shear angle range in degrees.
        '''
        self.shear_range = shear_range
        if shear_on:
            self.augmentation.add('shear')
        else:
            self.augmentation.discard('shear')

    def pixels(self):
        pix = self._normalize(self._pixels).astype(self.float_data_type)

        return np.moveaxis(pix, [0, 1, 2, 3, 4],
                           self.pixel_dimension_order)

    def weights(self):
        return np.moveaxis(self._weights, [0, 1, 2, 3, 4],
                           self.pixel_dimension_order)

    # @lru_cache(maxsize=10)
    def tile_positions(self, sliding=True):
        '''
        Get all possible sliding window tile positions of the dataset.

        Parameters
        ----------
        sliding: bool
            If True, sliding window positions are returned. Otherwise
            positions of non-overlapping tiles.

        Returns
        -------
        array_like
            List of 4-element tuples. Tuples define (image_id, z, x, y).
        '''

        tile_pos = []

        # compute shift of sliding window
        if sliding:
            shifted_zxy = np.ceil(np.array(self.tile_size_zxy)/3).astype('int')
            # overlap by one third
        else:
            shifted_zxy = None  # no overlap

        msg = 'Compute tile positions: '
        for i in progressbar(range(self.dataset.n_images), msg, 40):
            img_shape = self.dataset.image_dimensions(i)[1:]
            pos = [(i, p[0], p[1], p[2])
                   for p in compute_pos(img_shape,
                                        self.tile_size_zxy,
                                        sliding=shifted_zxy)]
            tile_pos += pos

        return tile_pos

    def _augment_params(self):
        '''
        select random augmentation parameters
        '''

        augment_params = {}

        if 'flip' in self.augmentation:
            _, x, y = self.tile_size_zxy
            is_square_tile = (x == y)

            augment_params = {'fliplr': np.random.choice([True, False]),
                              'flipud': np.random.choice([True, False]),
                              'rot90': np.random.choice(4)
                              if is_square_tile else 0}

        if 'rotate' in self.augmentation:
            augment_params['rotation_angle'] = random.uniform(
                *self.rotation_range)

        if 'shear' in self.augmentation:
            augment_params['shear_angle'] = random.uniform(*self.shear_range)

        return augment_params

    def remove_unlabeled_tiles(self):
        '''
        Scans all pixels and labels for all tiles. Removes all tile positions
        that do not contain labels.
        '''

        labels = np.array(sorted(self.labels))
        channels = np.array(sorted(self.channels))

        # Removing tiles from unlabeled slices
        if type(self.dataset.pixel_connector) == NapariConnector:
            self.tile_pos_for_label = self.effective_tiles()

        for label in labels:
            logger.info('scanning tiles for label {}...'.format(label))
            positions = self.tile_pos_for_label[label]
            n_pos = len(positions)

            for i, pos in enumerate(positions):
                image_nr = pos[0]
                pos_zxy = pos[1:]

                tile_data = self.dataset.training_tile(
                    image_nr,
                    pos_zxy,
                    self.tile_size_zxy,
                    channels,
                    labels,
                    pixel_padding=self.padding_zxy,
                    augment_params=None)

                if not _are_weights_in_tile(tile_data, label):
                    # remove tile position for the label, since no labels here
                    self.tile_pos_for_label[label].pop(i)
            n_pos_after = len(self.tile_pos_for_label[label])

            logger.info('removed {} tiles of {} for label {} ({}%)'.format(
                n_pos_after, n_pos, label, round(n_pos_after/n_pos*100., 2)))

    def split(self, fraction):
        '''
        Remove fraction of tiles from TrainingBatch and assign it to
        a new TrainingBatch. Can be used for separating training and
        validation data.
        Overlapping tile positions are removed from the parent TrainingBatch.

        Parameters
        ----------
        fraction: float
            Approx. fraction of data for the returned (child) TrainingBatch.
            Between 0 and 1.

        Returns
        -------
        array_like
            List of 4-element tuples. Tuples define (image_id, z, x, y).
        '''

        assert fraction >= 0 and fraction <= 1

        labels = np.array(sorted(self.labels))
        shape = [1, 0, 0, 0]  # (image, z, x, y)
        shape[1:] = self.tile_size_zxy

        tile_pos_for_label_out = {}

        for label in labels:
            logger.info(
                'splitting approximate fraction of {} for label {}'.format(
                    fraction, label))
            pos = self.tile_pos_for_label[label]
            n_pos = len(pos)
            n_pos_out = round(n_pos * fraction)

            assert n_pos > 0
            assert n_pos_out > 0

            curr_fraction = 0

            pos_out = []
            while curr_fraction < fraction:
                # select tile positions randomly
                choice = np.random.choice(range(len(pos)))
                a = pos.pop(choice)
                pos_out.append(a)

                is_overlap_ind = np.nonzero(find_overlapping_tiles(
                    a,
                    pos,
                    shape))[0]
                is_overlap_ind = sorted(is_overlap_ind, reverse=True)

                [pos.pop(ind) for ind in is_overlap_ind]

                curr_fraction = float(len(pos_out))/len(pos)

            tile_pos_for_label_out[label] = sorted(pos_out)
            self.tile_pos_for_label[label] = sorted(pos)

            tiles_remain = 100. * len(pos)/(len(pos) + len(pos_out))
            logger.info('remaining tiles: {} ({}%)'.format(
                len(pos), round(tiles_remain, 2)))

            tiles_removed = 100. * len(pos_out)/(len(pos) + len(pos_out))
            logger.info('removed tiles: {} ({}%)'.format(
                len(pos_out), round(tiles_removed, 2)))

        out = TrainingBatch(self.dataset,
                            self.tile_size_zxy,
                            padding_zxy=self.padding_zxy,
                            equalized=self.equalized)

        out.pixel_dimension_order = self.pixel_dimension_order
        out._batch_size = self._batch_size
        out.normalize_mode = self.normalize_mode
        out.global_norm_minmax = self.global_norm_minmax
        out.float_data_type = self.float_data_type
        out.equalized = self.equalized
        out.augmentation = self.augmentation
        out.rotation_range = self.rotation_range
        out.shear_range = self.shear_range

        out.tile_pos_for_label = tile_pos_for_label_out

        return out

    def _random_tile(self, for_label):
        '''
        Pick random tile in image regions where label data is present.
        '''

        # random pollng loop
        counter = 0
        while counter <= len(self.tile_pos_for_label[for_label]):
            counter += 1
            pos = self.tile_pos_for_label[for_label]

            msg = 'no label data for label {} in dataset'.format(for_label)
            assert len(pos) > 0, msg

            choice = np.random.randint(0, len(pos))
            pos_selected = pos[choice]

            image_nr = pos_selected[0]
            pos_zxy = pos_selected[1:]

            labels = np.array(sorted(self.labels))
            channels = np.array(sorted(self.channels))
            tile_data = self.dataset.training_tile(
                image_nr,
                pos_zxy,
                self.tile_size_zxy,
                channels,
                labels,
                pixel_padding=self.padding_zxy,
                augment_params=self._augment_params())

            if _are_weights_in_tile(tile_data, for_label):
                msg = ('Needed {} trials to fetch random tile containing ' +
                       'labelvalue {}').format(counter,
                                               for_label)
                logger.info(msg)

                return tile_data

            else:
                # remove tile position for the label, since no labels here
                r = self.tile_pos_for_label[for_label].pop(choice)

        msg = ('Could not fetch random tile containing labelvalue {} ' +
               'within {} trials').format(for_label, counter)
        logger.warning(msg)

        return tile_data


def _are_weights_in_tile(tile_data, for_label):
    '''
    check if weights for specified label are present
    '''
    lblregion_index = np.where(tile_data.labels == for_label)
    weights_lblregion = tile_data.weights[lblregion_index]
    return weights_lblregion.any()



import os

from setuptools import setup

reqs = ['numpy>=1.13.1',
        'munkres>=1.0.8',
        'scikit_image>=0.12.3',
        'tifffile',
        'h5py',
        'sparse>=0.12.0']


def readme():
    README_rst = os.path.join(os.path.dirname(__file__), 'README.md')
    with open(README_rst) as f:
        return f.read()


setup(name='yapic_io',
      version='0.2.6',
      description='io data handling module for various image sources as interface for pixel classification tools',
      long_description=readme(),
      long_description_content_type='text/markdown',
      author='Manuel Schoelling, Christoph Moehl',
      author_email='manuel.schoelling@dzne.de, christoph.moehl@dzne.de',
      packages=['yapic_io'],
      zip_safe=False,
      install_requires=reqs,
      test_suite='nose.collector',
      tests_require=['coverage', 'nose-timer', 'nose'])



import numpy as np
from numpy.random import choice, randint
import random
import collections
import yapic_io.utils as ut
from functools import lru_cache
import logging
import os
import yapic_io.transformations as trafo
import sys

logger = logging.getLogger(os.path.basename(__file__))
logger.setLevel(logging.INFO)
randint_array = np.vectorize(randint)
TrainingTile = collections.namedtuple('TrainingTile',
                                      ['pixels', 'channels', 'weights',
                                       'labels', 'augmentation'])


class Dataset(object):
    '''
    Provides connectors to pixel data source and assigned weights for
    classifier training.
    Provides methods for getting image tiles and data augmentation for
    classifier training, as well as writing classifier output tile-by-tile
    to target images.



    Parameters
    ----------
    pixel_connector : yapic_io.connector.Connector
        Connector object (e.g. TiffConnector or IlastikConnector) for binding
        of pixel and label data, as well as prediction result data.

    Notes
    -----
    Pixel data is loaded lazily to allow images of arbitrary size.
    Pixel data is cached in memory for repeated requests.
    '''

    def __init__(self, pixel_connector):

        self.pixel_connector = pixel_connector
        self.n_images = pixel_connector.image_count()
        self.label_counts = self.load_label_counts()

        # self.label_weights dict is complementary to self.label_counts
        self.label_weights = {label: 1 for label in self.label_counts.keys()}

        # max nr of trials to get a random training tile in polling mode
        self.max_pollings = 30

        is_consistent, channel_cnt = self.channels_are_consistent()
        msg = ('Varying number of channels: {}. '
               'Channel counts must be identical '
               'for all images in dataset. '
               'Dataset may be incomplete.').format(channel_cnt)
        assert is_consistent, msg

    def __repr__(self):
        return 'Dataset ({} images)'.format(self.n_images)

    @lru_cache(maxsize=1000)
    def image_dimensions(self, image_nr):
        '''
        Returns dimensions of the dataset.
        dims is a 4-element-tuple:

        Parameters
        ----------
        image_nr : int
            index of image

        Returns
        -------
        (nr_channels, nr_zslices, nr_x, nr_y)
        '''
        return self.pixel_connector.image_dimensions(image_nr)

    def _smallest_image_size_xy(self):
        Z = 10
        X = 500
        Y = 500
        for i in range(self.n_images):
            size_z, size_x, size_y = self.image_dimensions(i)[-3:]
            if size_z < Z:
                Z = size_z
            if size_x < X:
                X = size_x
            if size_y < Y:
                Y = size_y
        return (Z, X, Y)

    def pixel_statistics(self,
                         channels,
                         upper=99,
                         lower=1,
                         tile_size_zxy=None,
                         n_tiles=1000):
        '''
        Performs random sampling of n tiles and calculates upper and lower
        percentiles of pixel values. Separated for each channel.
        This data can be used for normalization of pixel intensities.

        Parameters
        ----------
        channels : array_like
            List of pixel channels to be fetched.
        upper : float
            Upper percentile.
        lower : float
            Lower percentile.
        tile_size_zxy : (nr_zslices, nr_x, nr_y)
            Tile size.
        n_tiles : int
            Nr of tiles to be fetched

        Returns
        -------
        [(lower_01, upper_01), (lower_02, upper_02), (lower_03, upper_03), ...]
        '''

        if tile_size_zxy is None:
            tile_size_zxy = self._smallest_image_size_xy()
        percentiles = np.zeros((n_tiles, len(channels), 2))
        msg = ('\n\nCalculate global pixel statistics'
               '({} tiles of size {})...\n').format(n_tiles, tile_size_zxy)
        sys.stdout.write(msg)
        for i in range(n_tiles):
            image_nr, pos_zxy = self._random_pos_izxy(None, tile_size_zxy)
            pix = self.multichannel_pixel_tile(image_nr,
                                               pos_zxy,
                                               tile_size_zxy,
                                               channels)
            percentiles[i, :, 0] = np.percentile(pix, lower, axis=[1, 2, 3])
            percentiles[i, :, 1] = np.percentile(pix, upper, axis=[1, 2, 3])

        lower_bnds = np.min(percentiles[:, :, 0], axis=0)
        upper_bnds = np.max(percentiles[:, :, 1], axis=0)

        out = [(lower, upper) for lower, upper in zip(lower_bnds, upper_bnds)]
        msg = '{} and {} percentiles per channel: {}\n'.format(lower,
                                                               upper,
                                                               out)
        sys.stdout.write(msg)
        return out

    def channels_are_consistent(self):
        '''
        Returns True if all images of the dataset have the same number of
        channels. Otherwise False.

        Returns
        -------
        boolean
            True if channels are consistent
        list
            List with channel counts
        '''
        channel_cnt = np.unique([self.image_dimensions(i)[0]
                                for i in range(self.n_images)])

        if len(channel_cnt) == 1:
            return True, channel_cnt

        return False, channel_cnt

    def label_values(self):
        '''
        Get label values.

        Returns
        -------
        list
            Sorted list of label values.
        '''
        labels = list(self.label_counts.keys())
        labels.sort()
        return labels

    def random_training_tile(self,
                             size_zxy,
                             channels,
                             pixel_padding=(0, 0, 0),
                             equalized=False,
                             augment_params=None,
                             labels='all',
                             ensure_labelvalue=None):
        '''
        Returns a randomly chosen training tile including weights.

        Parameters
        ----------
        size_zxy : (nr_zslices, nr_x, nr_y)
            Tile size.
        channels : array_like
            List of pixel channels to be fetched.
        pixel_padding : (pad_z, pad_x, pad_y)
            Amount of padding to increase tile size in zxy.
        equalized : bool
            If true, less frequent label_values are picked with same
            probability as frequent label_values.
        augment_params : dict
            Image augmentation settings. Possible key:value pairs:
            fliplr : True or False;
            flipud : True or False;
            rot90 : int, number of 90 degree rotations;
            rotate : float, rotation in degrees;
            shear : float, shear in degrees.
        labels : array_like or str
            List of labelvalues to be fetched.
        ensure_labelvalue : int
            Labelvalue that must be present in the fetched tile region.

        Returns
        -------
        collections.namedtuple
            TrainingTile(pixels, channels, labels, weights, augmentation)
        '''
        augment_params = augment_params or {}
        if labels == 'all':
            labels = self.label_values()

        if hasattr(self.pixel_connector, 'label_index_to_coordinate'):
            # fetch by label index
            return self._random_training_tile_by_coordinate(
                size_zxy,
                channels,
                labels,
                pixel_padding=pixel_padding,
                equalized=equalized,
                augment_params=augment_params,
                ensure_labelvalue=ensure_labelvalue)
        else:
            return self._random_training_tile_by_polling(
                size_zxy,
                channels,
                labels,
                pixel_padding=pixel_padding,
                equalized=equalized,
                augment_params=augment_params,
                ensure_labelvalue=ensure_labelvalue)

    def _random_pos_izxy(self, label_value, tile_size_zxy):
        '''
        Get a random image and a random zxy position (for a tile of shape
        size_zxy) within this image.
        Images with more frequent labels of type label_value are more likely to
        be selected.
        '''
        # get random image by label probability
        # label probability per image
        label_prob = self._get_label_probs(label_value)
        img_nr = choice(len(label_prob), p=label_prob)

        # get random zxy position within selected image
        img_shape_zxy = self.image_dimensions(img_nr)[1:]
        img_maxpos_zxy = np.array(img_shape_zxy) - tile_size_zxy

        msg = 'Tile of size {} does not fit in image of size {}'.format(
                    tile_size_zxy, img_shape_zxy)
        assert (img_maxpos_zxy > -1).all(), msg

        pos_zxy = randint_array(img_maxpos_zxy + 1)
        return img_nr, pos_zxy

    def _get_label_probs(self, label_value):
        '''
        Get probabilities for labels per image if label_value is None,
        probabilities for the sum of all label values is returned.
        '''
        if label_value is None:
            # label count for all labelvalues
            lbl_count = sum(self.label_counts.values())
        else:
            lbl_count = self.label_counts[label_value]
        return lbl_count / lbl_count.sum()

    def _random_training_tile_by_polling(self,
                                         size_zxy,
                                         channels,
                                         labels,
                                         pixel_padding=(0, 0, 0),
                                         equalized=False,
                                         augment_params=None,
                                         ensure_labelvalue=None):
        ''''
        Fetches a random training tile by repeated polling until the label
        specified in ensure_labelvalue is at least one time represented in the
        tile. The number if trials is set in self.max_pollings.
        If the nr of trials exceeds max_pollings, the last fetched tile is
        returned, although not containing the label.
        '''
        augment_params = augment_params or {}
        if ensure_labelvalue is None and equalized:
            ensure_labelvalue = self._random_label_value(equalized=equalized)

        for counter in range(self.max_pollings):
            img_nr, pos_zxy = self._random_pos_izxy(ensure_labelvalue,
                                                    size_zxy)

            tile_data = self.training_tile(img_nr, pos_zxy, size_zxy,
                                           channels, labels,
                                           pixel_padding=pixel_padding,
                                           augment_params=augment_params)

            if ensure_labelvalue is None:
                # check if weights for any label are present
                are_weights_in_tile = tile_data.weights.any()
            else:
                # convert set to sorted array
                labels = np.array(sorted(tile_data.labels))
                # check if weights for specified label are present
                lblregion_index = np.where(labels == ensure_labelvalue)[0]
                weights_lblregion = tile_data.weights[lblregion_index]
                are_weights_in_tile = weights_lblregion.any()

            if are_weights_in_tile:
                msg = ('Needed {} trials to fetch random tile containing ' +
                       'labelvalue {}').format(counter,
                                               ensure_labelvalue or '<any>')
                if counter == 0:
                    logger.debug(msg)
                else:
                    logger.info(msg)
                return tile_data

        msg = ('Could not fetch random tile containing labelvalue {} ' +
               'within {} trials').format(ensure_labelvalue, counter)
        logger.warning(msg)

        # if no labelweights are present from any labelvalue
        if not tile_data.weights.any():
            msg = ('training labelweighs do not contain any weights above 0 ' +
                   'for any label')
            logger.warning(msg)

        return tile_data

    def _random_training_tile_by_coordinate(self,
                                            size_zxy,
                                            channels,
                                            labels,
                                            pixel_padding=(0, 0, 0),
                                            equalized=False,
                                            augment_params=None,
                                            ensure_labelvalue=None):
        augment_params = augment_params or {}
        if ensure_labelvalue is None:
            ensure_labelvalue = self._random_label_value(equalized=equalized)

        total_count = self.label_counts[ensure_labelvalue].sum()
        img_nr, _, *pos_zxy = self.label_coordinate(ensure_labelvalue,
                                                    randint(total_count))
        np.testing.assert_array_equal(len(pos_zxy), len(size_zxy))

        # Now we have a label. But we can but the tile anywhere as long as the
        # label is in it

        pos_zxy = np.array(pos_zxy)
        shape_zxy = np.array(self.image_dimensions(img_nr)[1:])

        maxpos = np.minimum(pos_zxy, shape_zxy - size_zxy)
        minpos = np.maximum(0, pos_zxy - size_zxy + 1)

        pos_zxy = [np.random.randint(a, b + 1) for a, b in zip(minpos, maxpos)]

        tile_data = self.training_tile(img_nr, pos_zxy, size_zxy,
                                       channels, labels,
                                       pixel_padding=pixel_padding,
                                       augment_params=augment_params)

        return tile_data

    def training_tile(self,
                      image_nr,
                      pos_zxy,
                      size_zxy,
                      channels,
                      labels,
                      pixel_padding=(0, 0, 0),
                      augment_params=None):
        '''
        Returns a training tile including weights.

        Parameters
        ----------
        image_nr : int
            Index of image.
        pos_zxy : (z, x, y)
            Upper left position of pixels in source image_nr.
        size_zxy : (nr_zslices, nr_x, nr_y)
            Tile size.
        channels : array_like
            List of pixel channels to be fetched.
        labels : array_like
            List of labelvalues to be fetched.
        pixel_padding : (pad_z, pad_x, pad_y)
            Amount of padding to increase tile size in zxy.
        augment_params : dict
            Image augmentation settings. Possible key:value pairs:
            fliplr : True or False;
            flipud : True or False;
            rot90 : int, number of 90 degree rotations;
            rotate : float, rotation in degrees;
            shear : float, shear in degrees.

        Returns
        -------
        collections.namedtuple
            TrainingTile(pixels, channels, labels, weights, augmentation)
        '''
        augment_params = augment_params or {}
        # 4d pixel tile with selected channels in 1st dimension
        pixel_tile = self.multichannel_pixel_tile(
                        image_nr, pos_zxy, size_zxy, channels,
                        pixel_padding=pixel_padding,
                        augment_params=augment_params)

        # 4d label tile with selected labels in 1st dimension
        shape_zxy = self.image_dimensions(image_nr)[1:]
        label_tile = [_augment_tile(shape_zxy, pos_zxy, size_zxy,
                                    self._get_weights_tile,
                                    augment_params=augment_params,
                                    image_nr=image_nr,
                                    label_value=l)
                      for l in labels]
        label_tile = np.array(label_tile)

        msg = 'pixel tile dim={} label tile dim={} labels={}'.format(
                    pixel_tile.shape, label_tile.shape, len(labels))
        logger.debug(msg)

        return TrainingTile(pixel_tile, channels, label_tile, labels,
                            augment_params)

    def multichannel_pixel_tile(self,
                                image_nr,
                                pos_zxy,
                                size_zxy,
                                channels,
                                pixel_padding=(0, 0, 0),
                                augment_params=None):
        augment_params = augment_params or {}
        np.testing.assert_equal(len(pos_zxy), 3,
                                'Expected 3 dimensions (Z, X, Y)')
        np.testing.assert_equal(len(size_zxy), 3,
                                'Expected 3 dimensions (Z, X, Y)')

        image_shape_zxy = self.image_dimensions(image_nr)
        ut.assert_valid_image_subset(image_shape_zxy[1:], pos_zxy, size_zxy)

        pixel_padding = np.array(pixel_padding)
        size_padded = size_zxy + 2 * pixel_padding
        pos_padded = pos_zxy - pixel_padding

        for c in channels:
            msg = 'channel {} does not exist'.format(c)
            assert c < self.pixel_connector.image_dimensions(image_nr)[0], msg
        tile = [_augment_tile(image_shape_zxy,
                              np.hstack([[c], pos_padded]),
                              np.hstack([[1], size_padded]),
                              self.pixel_connector.get_tile,
                              augment_params=augment_params,
                              image_nr=image_nr)
                for c in channels]

        return np.vstack(tile)

    def _get_weights_tile(self, image_nr=None, pos=None, size=None,
                          label_value=None):
        '''
        Returns a 3d weight matrix tile for a certain label with
        dimensions zxy.
        '''
        assert label_value in self.label_values()

        boolmat = self.pixel_connector.label_tile(image_nr, pos, size,
                                                  label_value)

        weight_mat = np.zeros_like(boolmat, float)
        weight_mat[boolmat] = self.label_weights[label_value]

        return weight_mat

    def equalize_label_weights(self):
        '''
        equalizes labels according to their amount.
        less frequent labels are weighted higher than more frequent labels
        '''

        weights = {
            l: np.nan_to_num(1.0 / img_counts.sum())
            for l, img_counts in self.label_counts.items()
        }

        self.label_weights = {l: c / sum(weights.values())
                              for l, c in weights.items()}

    def load_label_counts(self):
        '''
        Returns the cout of each labelvalue for each image as dict

        Returns
        -------
        dict
            Keys are labelvalues. Values are lists with label counts for each
            image.

        Examples
        --------
        >>> from yapic_io.dataset import Dataset
        >>> from yapic_io.tiff_connector import TiffConnector
        >>> from pprint import pprint
        >>> pixel_image_dir = 'yapic_io/test_data/tiffconnector_1/im/*.tif'
        >>> label_image_dir = 'yapic_io/test_data/tiffconnector_1/labels/*.tif'
        >>>
        >>> c = TiffConnector(pixel_image_dir, label_image_dir)
        >>> d = Dataset(c)
        >>> d.load_label_counts()
        {1: array([4, 0, 0]), 2: array([ 3,  0, 11]), 3: array([3, 0, 3])}
        '''

        logger.debug('Starting to load label counts...')

        label_counts = collections.defaultdict(lambda: np.zeros(self.n_images,
                                                                dtype='int64'))
        for i in range(self.n_images):
            img_label_counts = \
                self.pixel_connector.label_count_for_image(i) or {}

            for label_value in img_label_counts.keys():
                label_counts[label_value][i] = img_label_counts[label_value]

        return dict(label_counts)

    def sync_label_counts(self, datset):
        '''
        Should be applied e.g. if two datasets were created from splitted
        conncetors( e.g. with tiff_connector.split(). Typical use case
        is syncing a training dataset and a validation dataset.

        Parameters
        ----------
        datset : Dataset
             Dataset to sync labeldata.

        Notes
        -----
        When splitting Connectors it may occur that one or more label values
        are only present in one Connector. Sync_label_counts adds missing
        label values to a dataset. Note that the missing label values will
        have a label count of 0 after syncing.

        Examples
        --------
        Split one TiffConnector in two and sync label data.

        >>> from yapic_io.dataset import Dataset
        >>> from yapic_io.tiff_connector import TiffConnector
        >>> from pprint import pprint
        >>>
        >>> pixel_image_dir = 'yapic_io/test_data/tiffconnector_1/im/*.tif'
        >>> label_image_dir = 'yapic_io/test_data/tiffconnector_1/labels/*.tif'
        >>> c = TiffConnector(pixel_image_dir, label_image_dir)
        >>> c1, c2 = c.split(1./3.)
        >>>
        >>> d = Dataset(c)
        >>> d1 = Dataset(c1)
        >>> d2 = Dataset(c2)
        >>>
        >>> pprint(d.label_counts) # full dataset has three labelvalues
        {1: array([4, 0, 0]), 2: array([ 3,  0, 11]), 3: array([3, 0, 3])}
        >>>
        >>> pprint(d1.label_counts) # d1 has also three labelvalues
        {1: array([4]), 2: array([3]), 3: array([3])}
        >>>
        >>> pprint(d2.label_counts) #d2 has only two labelvalues
        {2: array([ 0, 11]), 3: array([0, 3])}
        >>>
        >>>
        >>> d1.sync_label_counts(d2)
        >>> pprint(d1.label_counts)
        {1: array([4]), 2: array([3]), 3: array([3])}
        >>>
        >>> pprint(d2.label_counts) #labelvalue 1 added to d2 (with count 0)
        {1: array([0, 0]), 2: array([ 0, 11]), 3: array([0, 3])}

        '''
        lc1 = self.label_counts
        lc2 = datset.label_counts

        # missing label in each dataset
        missing_in_d1 = set(lc2.keys()) - set(lc1.keys())
        missing_in_d2 = set(lc1.keys()) - set(lc2.keys())

        for lbl in missing_in_d1:
            lc1[lbl] = np.zeros((self.n_images), dtype=np.int64)
        for lbl in missing_in_d2:
            lc2[lbl] = np.zeros((datset.n_images), dtype=np.int64)

    def _random_label_value(self, equalized=False):
        '''
        Returns a randomly chosen labelvalue.

        Parameters
        ----------
        equalized : bool, optional
            If true, less frequent label_values are picked with same
            probability as frequent label_values.
        '''
        if equalized:
            return random.choice(self.label_values())

        # probabilities for each labelvalue
        total_counts = [counts.sum() for counts in self.label_counts.values()]
        total_counts_norm = np.array(total_counts) / sum(total_counts)

        # pick a labelvalue according to the labelvalue probability
        return np.random.choice(self.label_values(), p=total_counts_norm)


def inner_tile_size(image_shape, pos, tile_shape):
    '''
    If a requested tile is out of bounds, this function calculates a transient
    tile position size and pos. The transient tile has to be padded in a later
    step to extend the edges for delivering the originally requested out of
    bounds tile. The padding sizes for this step are also delivered.
    size_out and pos_out that can be used in a second step with padding.
    pos_tile defines the position in the transient tile for cutting out the
    originally requested tile.

    Paramters
    ---------
    image_shape: array_like
        shape of full size original image
    pos: array_like
        Upper left position of tile in full size original image
    tile_shape: array_like
        Size of tile.

    Returns
    -------
    pos_out, size_out, pos_tile, padding_sizes
        pos_out is the position inside the full size original image for the
        transient tile.
        (more explanation needed)
    '''
    padding_lower = np.maximum(0, -pos)
    padding_upper = np.maximum(0, pos - (image_shape - tile_shape))
    padding = np.vstack([padding_lower, padding_upper]).T

    pos_tile = np.maximum(0, pos + padding_upper - image_shape)
    pos_out = pos + padding_lower - pos_tile

    # WTF is going on here??
    size_inmat = tile_shape + np.minimum(0, pos)

    size_new_1 = np.maximum(padding_lower, size_inmat)
    size_new_2 = np.minimum(tile_shape, image_shape - pos_out)
    size_out = np.minimum(size_new_1, size_new_2)

    return tuple(pos_out), tuple(size_out), tuple(pos_tile), tuple(padding)


def _augment_tile(img_shape,
                  pos,
                  tile_shape,
                  get_tile_func,
                  augment_params=None,
                  **kwargs):
    '''
    fetch tile and augment it
    if rotation and shear is activated, a 3 times larger tile
    is fetched and the final tile is cut out from that after
    rotation/shear.
    '''
    augment_params = augment_params or {}
    rotation_angle = augment_params.get('rotation_angle', 0)
    shear_angle = augment_params.get('shear_angle', 0)

    pos = np.array(pos)
    orig_tile_shape = np.array(tile_shape)
    tile_shape = np.array(tile_shape)

    augment_fast = (tile_shape[-2:] > 1).any()
    augment_slow = augment_fast and (rotation_angle > 0 or shear_angle > 0)

    if augment_slow:
        pos -= tile_shape
        tile_shape *= 3

    res = inner_tile_size(img_shape, pos, tile_shape)
    pos_transient, size_transient, pos_inside_transient, pad_size = res

    tile = get_tile_func(pos=pos_transient, size=size_transient, **kwargs)

    tile = np.pad(tile, pad_size, mode='symmetric')
    mesh = ut.get_tile_meshgrid(tile.shape, pos_inside_transient, tile_shape)
    tile = tile[tuple(mesh)]

    if augment_fast:
        # if the requested tile is only of size 1 in x and y,
        # augmentation can be omitted, since rotation and flipping always
        # occurs around the center axis.
        rot90 = augment_params.get('rot90', 0)
        flipud = augment_params.get('flipud', False)
        fliplr = augment_params.get('fliplr', False)

        tile = trafo.flip_image_2d_stack(tile, fliplr=fliplr,
                                         flipud=flipud, rot90=rot90)

    if augment_slow:
        tile = trafo.warp_image_2d_stack(tile, rotation_angle, shear_angle)
        mesh = ut.get_tile_meshgrid(tile.shape, orig_tile_shape,
                                    orig_tile_shape)
        tile = tile[tuple(mesh)]

    return tile



import numpy as np
import logging
import os
import itertools
from difflib import SequenceMatcher
from munkres import Munkres
import sys
logger = logging.getLogger(os.path.basename(__file__))


def progressbar(it, prefix="", size=60, file=sys.stdout):
    count = len(it)

    def show(j):
        x = int(size*j/count)
        file.write("%s[%s%s] %i/%i\r" % (prefix, "#"*x, "."*(size-x),
                   j,
                   count))
        file.flush()
    show(0)
    for i, item in enumerate(it):
        yield item
        show(i+1)
    file.write("\n")
    file.flush()


def get_tile_meshgrid(image_shape, pos, size):
    '''
    Returns coordinates for selection of a sub matrix, given a certain
    n-dimensional position (upper left) and n-dimensional size.

    Parameters
    ----------
    image_shape : tuple of size ndim
        Image shape.
    pos: tuple
        Upper left position of the tile in n dimensions.
    size : tuple of length ndim
        Size of tile.

    Returns
    -------
    array_like
        Multidimensional meshgrid defining the sub matrix
    '''
    assert_valid_image_subset(image_shape, pos, size)
    return [slice(p, p+s) for p, s in zip(pos, size)]


def assert_valid_image_subset(image_shape, pos, size):
    '''
    Check if a requested image subset defined by upper left pos and size
    fits into the image of given shape.

    Parameters
    ----------
    image_shape : tuple of size ndim
        Image shape.
    pos: tuple
        Upper left position of the tile in n dimensions.
    size : tuple of length ndim
        Size of tile.

    Raises
    ------
        AssertionError
    '''
    pos = np.array(pos)

    np.testing.assert_equal(len(image_shape), len(size))
    np.testing.assert_equal(len(image_shape), len(pos))
    assert not (pos < 0).any(), 'tile out of image bounds'
    assert not (image_shape < (pos + size)).any(), 'tile out of image bounds'


def compute_pos(img_shape, tile_shape, sliding=None):

    '''
    Computes all  positions for fetching tiles of a given size (tile
    window).

    If the tiles do not fit perfectly in the image, the last positions
    are corrected, such that the last and the second-last tile would
    have some overlap.

    Parameters
    ----------
    img_shape : tuple of size ndim
        Image shape.
    tile_shape : tuple of size ndim
        Tile shape.
    sliding : None or tuple
        if tuple, positions are calculated for overlapping tiles. The tile
        positions are shifted by  n pixels in each direction as defined in the
        tuple.
        If None, positions are calculated for non overlapping tiles. However
        the last tile could overlap with its neighbors.
    shifted : integer
        Nr of pixels to shift the sliding window in all dimensions.

    Returns
    -------
    array_like
        list of n-dimensional positions
    '''
    img_shape = np.asarray(img_shape)
    tile_shape = np.asarray(tile_shape)

    assert sliding is None or len(sliding) == len(img_shape)

    msg = 'tile size {} > image shape {}'.format(tile_shape, img_shape)
    assert not (tile_shape > img_shape).any(), msg

    pos_per_dim = [np.arange(0, length, step)
                   for length, step in zip(img_shape, tile_shape)]

    mod = (img_shape % tile_shape)  # nr of out of bounds pixels for last tile
    shift_last_tile = mod - tile_shape
    shift_last_tile[mod == 0] = 0

    for el, shift in zip(pos_per_dim, shift_last_tile):
        el[-1] = el[-1] + shift

    pos = list(itertools.product(*pos_per_dim))

    if sliding is None or len(pos) <= 1:
        # non overlapping tile positions
        return np.array(pos)

    # if sliding window: interpolate missing positions with mgrid
    n_dims = len(img_shape)

    slices = [slice(pos[0][i], pos[-1][i]+1, s)
              for s, i in zip(sliding, range(n_dims))]
    mesh = np.mgrid[slices]
    mesh = np.array(mesh).swapaxes(0, -1)

    n_dims = len(img_shape)
    pos_array = mesh.reshape((np.int(mesh.size/n_dims), n_dims))

    return pos_array


def find_overlapping_tiles(a, pos, shape):

    a = np.asarray(a)
    pos = np.asarray(pos)

    is_overlap = []
    for dim in range(pos.shape[-1]):
        is_overlap.append(
            np.stack([pos[:, dim] <= a[dim] + shape[dim]-1,
                      pos[:, dim] + shape[dim]-1 >= a[dim]]).all(axis=0))
    return np.stack(is_overlap).all(axis=0)


def segregate_tile_pos(pos, shape, choices):
    '''
    splits a vector of positions in two vectors and removes all overlapping
    posisions.
    '''
    pos = np.asarray(pos)

    p1 = np.delete(pos, choices, axis=0)
    p2 = pos[choices, :]

    for a in p2:
        is_overlap_ind = find_overlapping_tiles(a, p1, shape)
        p1 = np.delete(p1, np.nonzero(is_overlap_ind)[0], axis=0)

    return [tuple(e) for e in p1], [tuple(e) for e in p2]


def _compute_str_dist_matrix(s1, s2):
    '''
    - compute matrix of string distances for two lists of strings
    - normalize lengths of string lists (fill shorter list with empty strings)
    '''
    s1 = s1 + ['' for _ in range(len(s2) - len(s1))]
    s2 = s2 + ['' for _ in range(len(s1) - len(s2))]

    mat = np.zeros((len(s1), len(s2)))
    for i, a in enumerate(s1):
        for j, b in enumerate(s2):
            mat[i, j] = 1 - SequenceMatcher(None, a, b).ratio()

    return mat, s1, s2


def find_best_matching_pairs(s1, s2):
    '''
    Find global minimum for pairwise assignment of strings
    by using the munkres (hungarian) algorithmself.

    Parameters
    ----------
    s1 : array_like
        List of strings.
    s2 : array_like
       List of strings.

    Notes
    -----
    Counts of non-empty elements must be equal for s1 and s2.
    '''

    # remove empties
    s1 = list(filter(None, s1))
    s2 = list(filter(None, s2))

    if len(s1) == 0:
        assert len(s2) == 0
        return []

    mat, s1norm, s2norm = _compute_str_dist_matrix(s1, s2)

    # find assignment combination with lowest global cost
    m = Munkres()
    indexes = m.compute(mat)

    pairs = [[s1norm[i], s2norm[j]] for i, j in indexes]

    # change empty strings to None
    for pair in pairs:
        if pair[0] == '':
            pair[0] = None
        elif pair[1] == '':
            pair[1] = None

    return [pair for pair in pairs if pair[0] is not None]
